define(["exports", "core/pubsubhub", "core/symbols", "core/biblio-db"], function (exports, _pubsubhub, _symbols, _biblioDb) {
  "use strict";

  Object.defineProperty(exports, "__esModule", {
    value: true
  });
  exports.validateWebIDL = validateWebIDL;

  function _asyncToGenerator(fn) {
    return function () {
      var gen = fn.apply(this, arguments);
      return new Promise(function (resolve, reject) {
        function step(key, arg) {
          try {
            var info = gen[key](arg);
            var value = info.value;
          } catch (error) {
            reject(error);
            return;
          }

          if (info.done) {
            resolve(value);
          } else {
            return Promise.resolve(value).then(function (value) {
              step("next", value);
            }, function (err) {
              step("throw", err);
            });
          }
        }

        return step("next");
      });
    };
  }

  // Promise eventually gives us the URL for WebIDL from spec ref.
  const idlRefLookup = _biblioDb.biblioDB.find("WEBIDL").then(({ href }) => href);

  function validateWebIDL(parsedIdl) {
    for (const idl of parsedIdl) {
      switch (idl.type) {
        case "interface":
          validateInterface(idl);
          break;
        case "namespace":
          validateNamespace(idl);
          break;
        case "implements":
          validateImplements(idl);
          break;
      }
    }
    return parsedIdl;
  }

  function validateImplements(obj) {
    debugger;
    (0, _pubsubhub.pub)("The implements keyword is deprecated in Web IDL, use a 'mixin' instead.");
  }

  function validateNamespace(obj) {
    checkIfExplicitlyExposed(obj);
    checkForUndesirableExtAttrs(obj);
  }

  function validateInterface(obj) {
    checkIfExplicitlyExposed(obj);
    checkForUndesirableExtAttrs(obj);
  }

  /**
   * Checks for undesirable extended attributes on interfaces.
   *
   * @param {Object} obj Either a interface or a namespace as per
   * https://github.com/w3c/webidl2.js.
   */

  let checkForUndesirableExtAttrs = (() => {
    var _ref = _asyncToGenerator(function* (obj) {
      const { extAttrs } = obj;
      if (!extAttrs.length) {
        return;
      }
      const undesirables = extAttrs.filter(function ({ name }) {
        return _symbols.idlUndesirableExtAttrs.has(name);
      });
      if (!undesirables.length) {
        return;
      }
      for (const extAttr of undesirables) {
        const { name } = extAttr;
        const idlURL = yield idlRefLookup;
        const { name: owner } = extAttr.parent;
        const { type } = extAttr.parent;
        const msg = `Extended attribute \`[${name}]\` on \`${type}\` \`${owner}\`` + ` is an **undesirable feature**. [More info](${idlURL}#${name}).`;
        (0, _pubsubhub.pub)("warn", msg);
      }
    });

    return function checkForUndesirableExtAttrs(_x) {
      return _ref.apply(this, arguments);
    };
  })();

  let checkIfExplicitlyExposed = (() => {
    var _ref2 = _asyncToGenerator(function* (obj) {
      const exposed = obj.extAttrs.find(function ({ name }) {
        return _symbols.idlExposedKeywords.has(name);
      });
      if (!exposed) {
        return;
      }
      const hasRhs = exposed.arguments || exposed.rhs;
      if (hasRhs) {
        return;
      }
      // We have a bad one
      const { type, name } = obj;
      const idlURL = yield idlRefLookup;
      let msg = "";
      if (!exposed) {
        msg = `${type} \`${name}\` should declare an \`[Exposed=]\` extended attribute. `;
      } else {
        msg = `Expected ${type} \`${name}\` is \`[Exposed]\`,` + " but needs least one argument (e.g., `[Exposed=Window`). ";
      }
      msg += `[More info](${idlURL}#Exposed).`;
      (0, _pubsubhub.pub)("warn", msg);
    });

    return function checkIfExplicitlyExposed(_x2) {
      return _ref2.apply(this, arguments);
    };
  })();
});
//# sourceMappingURL=webidl-validator.js.map